/* Generated by Cython 0.29.32 */

#ifndef __PYX_HAVE__dpctl___sycl_event
#define __PYX_HAVE__dpctl___sycl_event

#include "Python.h"
struct Py_SyclEventObject;
struct PySyclEventObject;

/* "dpctl/_sycl_event.pxd":26
 * 
 * 
 * cdef public api class _SyclEvent [             # <<<<<<<<<<<<<<
 *     object Py_SyclEventObject,
 *     type Py_SyclEventType
 */
struct Py_SyclEventObject {
  PyObject_HEAD
  DPCTLSyclEventRef _event_ref;
};

/* "dpctl/_sycl_event.pxd":35
 * 
 * 
 * cdef public api class SyclEvent(_SyclEvent) [             # <<<<<<<<<<<<<<
 *     object PySyclEventObject,
 *     type PySyclEventType
 */
struct PySyclEventObject {
  struct Py_SyclEventObject __pyx_base;
  struct __pyx_vtabstruct_5dpctl_11_sycl_event_SyclEvent *__pyx_vtab;
};

#ifndef __PYX_HAVE_API__dpctl___sycl_event

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyTypeObject) Py_SyclEventType;
__PYX_EXTERN_C DL_IMPORT(PyTypeObject) PySyclEventType;

#endif /* !__PYX_HAVE_API__dpctl___sycl_event */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_sycl_event(void);
#else
PyMODINIT_FUNC PyInit__sycl_event(void);
#endif

#endif /* !__PYX_HAVE__dpctl___sycl_event */
