/* Generated by Cython 0.29.32 */

#ifndef __PYX_HAVE_API__dpctl___sycl_context
#define __PYX_HAVE_API__dpctl___sycl_context
#ifdef __MINGW64__
#define MS_WIN64
#endif
#include "Python.h"
#include "_sycl_context.h"

static PyTypeObject *__pyx_ptype_5dpctl_13_sycl_context__SyclContext = 0;
#define Py_SyclContextType (*__pyx_ptype_5dpctl_13_sycl_context__SyclContext)
static PyTypeObject *__pyx_ptype_5dpctl_13_sycl_context_SyclContext = 0;
#define PySyclContextType (*__pyx_ptype_5dpctl_13_sycl_context_SyclContext)

static DPCTLSyclContextRef (*__pyx_api_f_5dpctl_13_sycl_context_SyclContext_GetContextRef)(struct PySyclContextObject *) = 0;
#define SyclContext_GetContextRef __pyx_api_f_5dpctl_13_sycl_context_SyclContext_GetContextRef
static struct PySyclContextObject *(*__pyx_api_f_5dpctl_13_sycl_context_SyclContext_Make)(DPCTLSyclContextRef) = 0;
#define SyclContext_Make __pyx_api_f_5dpctl_13_sycl_context_SyclContext_Make
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif


static int import_dpctl___sycl_context(void) {
  PyObject *module = 0;
  module = PyImport_ImportModule("dpctl._sycl_context");
  if (!module) goto bad;
  if (__Pyx_ImportFunction(module, "SyclContext_GetContextRef", (void (**)(void))&__pyx_api_f_5dpctl_13_sycl_context_SyclContext_GetContextRef, "DPCTLSyclContextRef (struct PySyclContextObject *)") < 0) goto bad;
  if (__Pyx_ImportFunction(module, "SyclContext_Make", (void (**)(void))&__pyx_api_f_5dpctl_13_sycl_context_SyclContext_Make, "struct PySyclContextObject *(DPCTLSyclContextRef)") < 0) goto bad;
  __pyx_ptype_5dpctl_13_sycl_context__SyclContext = __Pyx_ImportType(module, "dpctl._sycl_context", "_SyclContext", sizeof(struct Py_SyclContextObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5dpctl_13_sycl_context__SyclContext) goto bad;
  __pyx_ptype_5dpctl_13_sycl_context_SyclContext = __Pyx_ImportType(module, "dpctl._sycl_context", "SyclContext", sizeof(struct PySyclContextObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5dpctl_13_sycl_context_SyclContext) goto bad;
  Py_DECREF(module); module = 0;
  return 0;
  bad:
  Py_XDECREF(module);
  return -1;
}

#endif /* !__PYX_HAVE_API__dpctl___sycl_context */
